# EasySQLite

**Version: 0.1.0**

A lightweight, intuitive Python library that acts as a wrapper around Python's built-in `sqlite3` module. Its primary purpose is to significantly simplify common database operations for users who are not familiar with SQL syntax or find the standard `sqlite3` API verbose for simple tasks.

## Problem Solved

Working directly with `sqlite3` requires SQL knowledge and careful handling of connections, cursors, transactions, and data types. EasySQLite provides a Pythonic interface, reducing boilerplate and abstracting common SQL commands.

## Features (v0.1)

* Connect to/create SQLite database files easily.
* Context manager (`with` statement) for automatic connection handling (commit/rollback, close).
* Conforms with the Python ideology, as this is more **python like** than using sql queries.
* List and delete database files (with confirmation).
* Table CRUD: Create, List, Describe, Rename, Delete (with confirmation/force flag).
* Column CRUD (via `ALTER TABLE`): Add, Rename\*, Delete\* (\*requires modern SQLite versions).
* Row CRUD: Add single/multiple rows (using dictionaries), Get rows with filtering/ordering/limiting, Update rows, Delete rows (with safety checks), Count rows.
* Simplified `JOIN` operations (INNER, LEFT, CROSS).
* Execute arbitrary SQL queries safely using parameterization.
* Basic logging for operations and errors.
* Type hinting for better developer experience.

## Installation

```bash
pip install easysqlite-lib 
````

## Quickstart

```python
from easysqlite import EasySQLite, EasySQLiteError

DB_FILE = 'my_app_data.db'

try:
    # Use the context manager for safe connection handling
    with EasySQLite(DB_FILE) as db:
        # Create a table if it doesn't exist
        db.create_table(
            'users',
            {'id': 'INTEGER PRIMARY KEY', 'name': 'TEXT', 'email': 'TEXT UNIQUE'}
        )

        # Add some data
        user_id = db.add_row('users', {'name': 'Alice', 'email': 'alice@example.com'})
        db.add_rows('users', [
            {'name': 'Bob', 'email': 'bob@example.com'},
            {'name': 'Charlie', 'email': 'charlie@example.com'}
        ])

        # Retrieve data
        print("All users:")
        all_users = db.get_rows('users')
        for user in all_users:
            print(user) # Rows are dict-like

        print("\nUser with ID 1:")
        user1 = db.get_rows('users', condition={'id': 1})
        print(user1)

        # Update data
        db.update_rows('users', {'email': 'robert@example.com'}, condition={'name': 'Bob'})
        print("\nBob's updated record:")
        print(db.get_rows('users', condition={'name': 'Bob'}))

        # Count rows
        print(f"\nTotal users: {db.count_rows('users')}")

        # Delete data
        db.delete_rows('users', condition={'name': 'Charlie'})
        print(f"\nTotal users after delete: {db.count_rows('users')}")

        # Custom query (safe with parameters)
        print("\nUsers with example.com emails (custom query):")
        result = db.execute_query("SELECT name FROM users WHERE email LIKE ?", ('%@example.com',))
        if result['success']:
            print(result['data'])

except EasySQLiteError as e:
    print(f"Database operation failed: {e}")
except Exception as e:
    print(f"An unexpected error occurred: {e}")

```

## Error Handling

The library primarily uses return values (like `bool`, `int`, `List`, `Optional[int]`) to indicate success or simple results. For more significant errors during operations (e.g., SQL syntax errors, constraint violations, non-existent tables/columns where not expected, invalid input), it raises custom exceptions derived from `EasySQLiteError` (e.g., `TableError`, `RowError`, `QueryError`). Check the specific method docstrings for details on potential exceptions.

## Security

  * **SQL Injection:** The library uses parameterization (`?` placeholders) internally for all methods that accept data values (`add_row`, `add_rows`, `get_rows`, `update_rows`, `delete_rows`, `execute_query` with `params`). This protects against SQL injection *for data values*.
  * **Identifiers:** Table names and column names provided by the user are generally assumed to be safe (developer-provided). **Do not construct table or column names directly from untrusted external input**, as this library does not sanitize identifiers.
  * **`execute_query`:** Be extremely careful when using `execute_query`. If you build the SQL string dynamically using external input *without* using the `params` argument for that input, you risk SQL injection.

## Limitations & SQLite Versions

  * **RENAME COLUMN:** Requires SQLite `3.25.0+`. Raises `NotImplementedError` on older versions.
  * **DROP COLUMN:** Requires SQLite `3.35.0+`. Raises `NotImplementedError` on older versions. The complex copy/recreate workaround is *not* implemented.
  * **JOINs:** The `on` condition string in `join_rows` must be provided correctly, including necessary table prefixes (`table.column`). The library does not parse or automatically qualify these.
  * **Error Handling Strategy:** Uses a mix of return values and custom exceptions. See method docstrings.
  * **Concurrency:** Uses `check_same_thread=False` for basic multi-threaded use cases but doesn't provide advanced concurrency control beyond standard SQLite capabilities. Not suitable for high-concurrency server applications without careful design.

## Contributing

Contributions are welcome\! Please open an issue or submit a pull request on GitHub. (Link to repo)

## License

This project is licensed under the Apache License, Version 2.0 - see the [LICENSE](LICENSE) file for details.

---
Apache License
                           Version 2.0, January 2004
                        http://www.apache.org/licenses/

   TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION

   1. Definitions.

      "License" shall mean the terms and conditions for use, reproduction,
      and distribution as defined by Sections 1 through 9 of this document.

      "Licensor" shall mean the copyright owner or entity authorized by
      the copyright owner that is granting the License.

      "Legal Entity" shall mean the union of the acting entity and all
      other entities that control, are controlled by, or are under common
      control with that entity. For the purposes of this definition,
      "control" means (i) the power, direct or indirect, to cause the
      direction or management of such entity, whether by contract or
      otherwise, or (ii) ownership of fifty percent (50%) or more of the
      outstanding shares, or (iii) beneficial ownership of such entity.

      "You" (or "Your") shall mean an individual or Legal Entity
      exercising permissions granted by this License.

      "Source" form shall mean the preferred form for making modifications,
      including but not limited to software source code, documentation
      source, and configuration files.

      "Object" form shall mean any form resulting from mechanical
      transformation or translation of a Source form, including but
      not limited to compiled object code, generated documentation,
      and conversions to other media types.

      "Work" shall mean the work of authorship, whether in Source or
      Object form, made available under the License, as indicated by a
      copyright notice that is included in or attached to the work
      (an example is provided in the Appendix below).

      "Derivative Works" shall mean any work, whether in Source or Object
      form, that is based on (or derived from) the Work and for which the
      editorial revisions, annotations, elaborations, or other modifications
      represent, as a whole, an original work of authorship. For the purposes
      of this License, Derivative Works shall not include works that remain
      separable from, or merely link (or bind by name) to the interfaces of,
      the Work and Derivative Works thereof.

      "Contribution" shall mean any work of authorship, including
      the original version of the Work and any modifications or additions
      to that Work or Derivative Works thereof, that is intentionally
      submitted to Licensor for inclusion in the Work by the copyright owner
      or by an individual or Legal Entity authorized to submit on behalf of
      the copyright owner. For the purposes of this definition, "submitted"
      means any form of electronic, verbal, or written communication sent
      to the Licensor or its representatives, including but not limited to
      communication on electronic mailing lists, source code control systems,
      and issue tracking systems that are managed by, or on behalf of, the
      Licensor for the purpose of discussing and improving the Work, but
      excluding communication that is conspicuously marked or otherwise
      designated in writing by the copyright owner as "Not a Contribution."

      "Contributor" shall mean Licensor and any individual or Legal Entity
      on behalf of whom a Contribution has been received by Licensor and
      subsequently incorporated within the Work.

   2. Grant of Copyright License. Subject to the terms and conditions of
      this License, each Contributor hereby grants to You a perpetual,
      worldwide, non-exclusive, no-charge, royalty-free, irrevocable
      copyright license to reproduce, prepare Derivative Works of,
      publicly display, publicly perform, sublicense, and distribute the
      Work and such Derivative Works in Source or Object form.

   3. Grant of Patent License. Subject to the terms and conditions of
      this License, each Contributor hereby grants to You a perpetual,
      worldwide, non-exclusive, no-charge, royalty-free, irrevocable
      (except as stated in this section) patent license to make, have made,
      use, offer to sell, sell, import, and otherwise transfer the Work,
      where such license applies only to those patent claims licensable
      by such Contributor that are necessarily infringed by their
      Contribution(s) alone or by combination of their Contribution(s)
      with the Work to which such Contribution(s) was submitted. If You
      institute patent litigation against any entity (including a
      cross-claim or counterclaim in a lawsuit) alleging that the Work
      or a Contribution incorporated within the Work constitutes direct
      or contributory patent infringement, then any patent licenses
      granted to You under this License for that Work shall terminate
      as of the date such litigation is filed.

   4. Redistribution. You may reproduce and distribute copies of the
      Work or Derivative Works thereof in any medium, with or without
      modifications, and in Source or Object form, provided that You
      meet the following conditions:

      (a) You must give any other recipients of the Work or
          Derivative Works a copy of this License; and

      (b) You must cause any modified files to carry prominent notices
          stating that You changed the files; and

      (c) You must retain, in the Source form of any Derivative Works
          that You distribute, all copyright, patent, trademark, and
          attribution notices from the Source form of the Work,
          excluding those notices that do not pertain to any part of
          the Derivative Works; and

      (d) If the Work includes a "NOTICE" text file as part of its
          distribution, then any Derivative Works that You distribute must
          include a readable copy of the attribution notices contained
          within such NOTICE file, excluding those notices that do not
          pertain to any part of the Derivative Works, in at least one
          of the following places: within a NOTICE text file distributed
          as part of the Derivative Works; within the Source form or
          documentation, if provided along with the Derivative Works; or,
          within a display generated by the Derivative Works, if and
          wherever such third-party notices normally appear. The contents
          of the NOTICE file are for informational purposes only and
          do not modify the License. You may add Your own attribution
          notices within Derivative Works that You distribute, alongside
          or as an addendum to the NOTICE text from the Work, provided
          that such additional attribution notices cannot be construed
          as modifying the License.

      You may add Your own copyright statement to Your modifications and
      may provide additional or different license terms and conditions
      for use, reproduction, or distribution of Your modifications, or
      for any such Derivative Works as a whole, provided Your use,
      reproduction, and distribution of the Work otherwise complies with
      the conditions stated in this License.

   5. Submission of Contributions. Unless You explicitly state otherwise,
      any Contribution intentionally submitted for inclusion in the Work
      by You to the Licensor shall be under the terms and conditions of
      this License, without any additional terms or conditions.
      Notwithstanding the above, nothing herein shall supersede or modify
      the terms of any separate license agreement you may have executed
      with Licensor regarding such Contributions.

   6. Trademarks. This License does not grant permission to use the trade
      names, trademarks, service marks, or product names of the Licensor,
      except as required for reasonable and customary use in describing the
      origin of the Work and reproducing the content of the NOTICE file.

   7. Disclaimer of Warranty. Unless required by applicable law or
      agreed to in writing, Licensor provides the Work (and each
      Contributor provides its Contributions) on an "AS IS" BASIS,
      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
      implied, including, without limitation, any warranties or conditions
      of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
      PARTICULAR PURPOSE. You are solely responsible for determining the
      appropriateness of using or redistributing the Work and assume any
      risks associated with Your exercise of permissions under this License.

   8. Limitation of Liability. In no event and under no legal theory,
      whether in tort (including negligence), contract, or otherwise,
      unless required by applicable law (such as deliberate and grossly
      negligent acts) or agreed to in writing, shall any Contributor be
      liable to You for damages, including any direct, indirect, special,
      incidental, or consequential damages of any character arising as a
      result of this License or out of the use or inability to use the
      Work (including but not limited to damages for loss of goodwill,
      work stoppage, computer failure or malfunction, or any and all
      other commercial damages or losses), even if such Contributor
      has been advised of the possibility of such damages.

   9. Accepting Warranty or Additional Liability. While redistributing
      the Work or Derivative Works thereof, You may choose to offer,
      and charge a fee for, acceptance of support, warranty, indemnity,
      or other liability obligations and/or rights consistent with this
      License. However, in accepting such obligations, You may act only
      on Your own behalf and on Your sole responsibility, not on behalf
      of any other Contributor, and only if You agree to indemnify,
      defend, and hold each Contributor harmless for any liability
      incurred by, or claims asserted against, such Contributor by reason
      of your accepting any such warranty or additional liability.

   END OF TERMS AND CONDITIONS

   APPENDIX: How to apply the Apache License to your work.

      To apply the Apache License to your work, attach the following
      boilerplate notice, with the fields enclosed by brackets "[]"
      replaced with your own identifying information. (Don't include
      the brackets!)  The text should be enclosed in the appropriate
      comment syntax for the file format. We also recommend that a
      file or class name and description of purpose be included on the
      same "printed page" as the copyright notice for easier
      identification within third-party archives.

   Copyright 2025 Arindam Singh

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

---
